#![no_std]
#![no_main]
#![feature(type_alias_impl_trait)]

use defmt::*;
use embassy_executor::Spawner;
use embassy_futures::select::select;
use embassy_futures::select::Either::*;
use embassy_rp::gpio;
use embassy_time::{Duration, Timer};
use embassy_rp::gpio::{Input, Level, Output, Pull, Flex};
use embassy_rp::i2c;
use embassy_rp::interrupt;
use embedded_hal_async::i2c::I2c;
use embassy_rp::peripherals::{UART0, I2C0};
use embassy_rp::uart::{self, UartRx, Parity};
use {defmt_rtt as _, panic_probe as _};
use heapless::String;

#[embassy_executor::task]
async fn reader(mut rx: UartRx<'static, UART0, uart::Async>) {
    info!("Reading...");
    loop {
        let mut buf = [0; 1];
        match rx.read(&mut buf).await {
            Result::Ok(()) => info!("RX {:?}", buf),
            Result::Err(e) => info!("RX error {:?}", e),
        }
    }
}

fn append_hex1<const N : usize>(s : &mut String<N>, num : u8) -> Result<(), ()> {
    let ch = if num < 10 {
        ('0' as u8 + num) as char
    } else {
        ('a' as u8 + num - 10) as char
    };
    s.push(ch)
}

fn append_hex2<const N : usize>(s : &mut String<N>, num : u8) -> Result<(), ()> {
    append_hex1(s, num >> 4)?;
    append_hex1(s, num & 0xf)
}

#[derive(Copy, Clone, Debug)]
#[cfg_attr(feature = "defmt", derive(defmt::Format))]
enum ScanResult {
    NAck,
    Ack,
    Reserved,
    Error
}

fn format_scan_result<const N : usize>(msg : &mut String<N>, scan_result : [ScanResult; 128]) -> Result<(), ()> {
    const MAX_ADDRESS : u8 = 128;
    for i in 0..MAX_ADDRESS/16 {
        let base = i*16;
        append_hex2(msg, base)?;
        msg.push_str(":")?;
        for j in 0..16 {
            msg.push_str(" ")?;
            match scan_result[(base+j) as usize] {
                ScanResult::Ack => append_hex2(msg, base+j)?,
                ScanResult::NAck => msg.push_str("--")?,
                ScanResult::Reserved => msg.push_str("rr")?,
                ScanResult::Error => msg.push_str("!!")?,
            }
        }
        msg.push_str("\n")?;
    }
    Ok(())
}


// adapted from arduino-pico:
// https://github.com/earlephilhower/arduino-pico/blob/6e52b72523b11470b2ad2b0578ca1500be238108/libraries/Wire/src/Wire.cpp#L257
// This seems to be the only way to do zero-length writes:
// see https://github.com/raspberrypi/pico-sdk/issues/238
async fn handle_clock_stretching<'d, SCL: gpio::Pin>(scl: &mut Flex<'d, SCL>) -> Result<(), i2c::Error> {
    use i2c::Error::*;
    use i2c::AbortReason::*;

    if scl.is_high() {
        return Ok(())
    }

    let is_high = scl.wait_for_high();
    let timeout = Timer::after(Duration::from_micros(100));

    return match select(is_high, timeout).await {
        First(_) => Ok(()),
        Second(_) => Err(Abort(ArbitrationLoss)),
    }
}
async fn i2c_write_zero_bytes_middle_part<'d, SCL: gpio::Pin, SDA: gpio::Pin, const FREQ : u64>(
        mut addr: u16, delay: Duration, scl: &mut Flex<'d, SCL>, sda: &mut Flex<'d, SDA>)
        -> Result<(), i2c::Error> {
    sda.set_as_input();
    Timer::after(delay).await;
    scl.set_as_input();
    handle_clock_stretching(scl).await?;
    sda.set_as_output();
    Timer::after(delay).await;
    scl.set_as_output();
    Timer::after(delay).await;
    for _ in 0..8 {
        addr <<= 1;
        sda.set_level(if (addr & (1 << 7)) != 0 { gpio::Level::High } else { gpio::Level::Low });
        Timer::after(delay).await;
        scl.set_as_input();
        Timer::after(delay).await;
        handle_clock_stretching(scl).await?;
        scl.set_as_output();
        Timer::after(Duration::from_micros(5)).await; // Ensure we don't change too close to clock edge
    }

    sda.set_as_input();
    Timer::after(delay).await;
    scl.set_as_input();
    handle_clock_stretching(scl).await?;

    let result = if sda.is_low() {
        Ok(())
    } else {
        use i2c::Error::*;
        use i2c::AbortReason::*;
        Err(Abort(NoAcknowledge))
    };
    Timer::after(delay).await;
    scl.set_as_output();

    result
}
async fn i2c_write_zero_bytes<SCL: gpio::Pin, SDA: gpio::Pin, const FREQ : u64>(addr: u16, scl: &mut SCL, sda: &mut SDA)
        -> Result<(), i2c::Error> {
    use i2c::Error::*;
    use i2c::AbortReason::*;

    let delay = Duration::from_micros((1000000 / FREQ) / 2);

    let mut sda = Flex::new(sda);
    let mut scl = Flex::new(scl);
    sda.set_as_input();
    sda.set_pull(Pull::Up);
    scl.set_as_input();
    scl.set_pull(Pull::Up);
    // We only need low or pullup so we can set the value to low and only toggle direction.
    // This is how OutputOpenDrain does it (and the only reason we use Flex is because we want the pullup).
    sda.set_low();
    scl.set_low();

    Timer::after(delay).await;

    // abort if SCL is stuck low
    handle_clock_stretching(&mut scl).await?;

    // handle SDA being stuck low
    // (This can happen for very simple I2C devices that only use SCL and don't have any other clocks that
    // they can use to handle timeouts, e.g. simple port expanders.)
    if sda.is_low() {
        // I think I have read that DW_apb_i2c does nine clocks. This seems resonable so let's do the same.
        for _ in 0..9 {
            scl.set_as_output();
            Timer::after(delay).await;
            scl.set_as_input();
            Timer::after(delay).await;

            if sda.is_high() {
                break;
            }
        }

        if sda.is_low() || scl.is_low() {
            // There is isn't much that we can do here. A start condition would reset the state but we cannot
            // generate it while one of the lines is stuck low.
            return Err(Abort(ArbitrationLoss))
        }

        // SDA is high, again. Good. We will generate a start condition soon. This should abort any
        // partial transfers that some devices might think are ongoing. We have already waited for half
        // a clock so we are good to go.
    }

    let result = i2c_write_zero_bytes_middle_part::<_, _, FREQ>(addr, delay, &mut scl, &mut sda).await;

    Timer::after(delay).await;
    sda.set_as_output();
    Timer::after(delay).await;
    scl.set_as_input();
    Timer::after(delay).await;
    sda.set_as_input();
    Timer::after(delay).await;

    return result;
}

#[embassy_executor::task]
async fn i2c_task(mut i2c_peri: embassy_rp::peripherals::I2C0,
        mut scl: embassy_rp::peripherals::PIN_13, mut sda: embassy_rp::peripherals::PIN_12,
        mut i2c_irq: embassy_rp::interrupt::I2C0_IRQ, i2c_config: i2c::Config) {
    use i2c::Error::*;
    use i2c::AbortReason::*;

    loop {
        if true {
            let mut i2c = i2c::I2c::new_async(&mut i2c_peri, &mut scl, &mut sda, &mut i2c_irq, i2c_config);
        }

        const MAX_ADDRESS : u8 = 128;
        let mut scan_result = [ScanResult::Error; MAX_ADDRESS as usize];
        for i in 0..MAX_ADDRESS {
            scan_result[i as usize] = match i2c_write_zero_bytes::<_, _, 400_000>(i as u16, &mut scl, &mut sda).await {
                Result::Ok(()) => ScanResult::Ack,
                Result::Err(Abort(NoAcknowledge)) => ScanResult::NAck,
                Result::Err(AddressReserved(_)) => ScanResult::Reserved,
                Result::Err(e) => {
                    warn!("I2C problem: {:02x}: {:?}", i, e);
                    ScanResult::Error
                }
            }
        }

        const MSG_LEN : usize = (MAX_ADDRESS as usize)*3 + (MAX_ADDRESS as usize)/16*6;
        let mut msg = String::<MSG_LEN>::new();
        match format_scan_result(&mut msg, scan_result) {
            Ok(()) => info!("I2C scan result:\n{}", msg),
            Err(()) => info!("I2C scan result: too long for our buffer!"),
        }

        Timer::after(Duration::from_secs(1)).await;
    }
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let p = embassy_rp::init(Default::default());
    info!("starting");

    // pinout:
    let mut _drive1 = Output::new(p.PIN_0, Level::Low);
    let mut _drive2 = Output::new(p.PIN_1, Level::Low);
    //let LED_W = p.PIN_2;
    let _ws2812_display = p.PIN_2;
    //let LED_Y = p.PIN_3;
    let mut _en_measure_current = Output::new(p.PIN_3, Level::Low);
    let mut led_b = Output::new(p.PIN_4, Level::Low);
    let mut _led_g = Output::new(p.PIN_5, Level::Low);
    let mut _led_r = Output::new(p.PIN_6, Level::Low);
    let _matrix_in1 = Input::new(p.PIN_7, Pull::Up);
    let _sbu1 = p.PIN_8;
    let _sbu2 = p.PIN_9;
    let _vbus_det = p.PIN_10;
    let _button_boot2 = Input::new(p.PIN_11, Pull::Up);
    let sda = p.PIN_12;
    let scl = p.PIN_13;
    let _ws2811_red_in = Input::new(p.PIN_14, Pull::None);
    let _tx_en = p.PIN_15;
    let _tx = p.PIN_16;
    let rx = p.PIN_17;
    let _matrix_in2 = Input::new(p.PIN_18, Pull::Up);
    let _reed3 = Input::new(p.PIN_19, Pull::Up);
    let _reed4 = Input::new(p.PIN_20, Pull::Up);
    let _reed2 = Input::new(p.PIN_21, Pull::Up);
    let _reed1 = Input::new(p.PIN_22, Pull::Up);
    let _matrix_in3 = Input::new(p.PIN_23, Pull::Up);
    let _digout1 = p.PIN_24;
    let _digout2 = p.PIN_25;
    let _analog_in1 = p.PIN_26;
    let _current2 = p.PIN_27;
    let _current1 = p.PIN_28;
    let _measure_vcc = p.PIN_29;

    info!("set up i2c");
    let i2c_irq = interrupt::take!(I2C0_IRQ);
    let mut i2c_config = i2c::Config::default();
    i2c_config.frequency = 400_000;
    unwrap!(spawner.spawn(i2c_task(p.I2C0, scl, sda, i2c_irq, i2c_config)));

    // use 19200 baud in 8E1 mode - not great but Modbus default
    let mut uart_config = uart::Config::default();
    uart_config.baudrate = 19200;
    uart_config.parity = Parity::ParityEven;
    let uart_rx = UartRx::new(
        p.UART0,
        rx,
        interrupt::take!(UART0_IRQ),
        p.DMA_CH1,
        uart_config,
    );
    unwrap!(spawner.spawn(reader(uart_rx)));

    loop {
        info!("led on!");
        led_b.set_high();
        Timer::after(Duration::from_secs(1)).await;

        info!("led off!");
        led_b.set_low();
        Timer::after(Duration::from_secs(1)).await;
    }
}
